#!/usr/bin/env python3
# Copyright (C) 2020-2024 All rights reserved.
# FILENAME:    ~~/src/fastapi_csrf_protect/csrf_config.py
# VERSION:     0.3.7
# CREATED:     2020-11-25 14:35
# AUTHOR:      Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard Packages ###
from typing import Callable, List, Literal, Optional, Set

### Third-Party Modules ###
from pydantic import ValidationError
from pydantic_settings import BaseSettings

### Local Modules ###
from fastapi_csrf_protect.load_config import LoadConfig


class CsrfConfig(object):
  _cookie_key: str = "fastapi-csrf-token"
  _cookie_path: str = "/"
  _cookie_domain: Optional[str] = None
  _cookie_samesite: Optional[Literal["lax", "strict", "none"]] = None
  _cookie_secure: bool = False
  _header_name: str = "X-CSRF-Token"
  _header_type: Optional[str] = None
  _httponly: bool = True
  _max_age: int = 3600
  _methods: Set[str] = {"POST", "PUT", "PATCH", "DELETE"}
  _secret_key: Optional[str] = None
  _token_location: str = "header"
  _token_key: str = "csrf-token"

  @classmethod
  def load_config(cls, settings: Callable[..., List[tuple] | BaseSettings]) -> None:
    try:
      config = LoadConfig(**{key.lower(): value for key, value in settings()})
      cls._cookie_key = config.cookie_key or cls._cookie_key
      cls._cookie_path = config.cookie_path or cls._cookie_path
      cls._cookie_domain = config.cookie_domain
      cls._cookie_samesite = config.cookie_samesite  # type: ignore[assignment]
      cls._cookie_secure = config.cookie_secure or cls._cookie_secure
      cls._header_name = config.header_name or cls._header_name
      cls._header_type = config.header_type
      cls._httponly = config.httponly or cls._httponly
      cls._max_age = config.max_age or cls._max_age
      cls._methods = config.methods or cls._methods
      cls._secret_key = config.secret_key
      cls._token_location = config.token_location or cls._token_location
      cls._token_key = config.token_key
    except ValidationError:
      raise
    except Exception as err:
      print(err)
      raise TypeError('CsrfConfig must be pydantic "BaseSettings" or list of tuple')


__all__ = ("CsrfConfig",)



#!/usr/bin/env python3
# Copyright (C) 2020-2024 All rights reserved.
# FILENAME:    ~~/tests/__init__.py
# VERSION:     0.3.7
# CREATED:     2020-11-26 18:50
# AUTHOR:      Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************
### Standard Packages ###
from pytest import fixture

### Third-Party Packages ###
from fastapi import Depends, FastAPI, Request
from fastapi.responses import JSONResponse
from fastapi.testclient import TestClient

### Local Modules ###
from fastapi_csrf_protect import CsrfProtect
from fastapi_csrf_protect.exceptions import CsrfProtectError


@fixture
def test_client() -> TestClient:
  """
  Sets up a FastAPI TestClient wrapped around an App implementing Context/Headers extension pattern

  ---
  :returns: TestClient
  """
  app = FastAPI()

  @app.get("/gen-token", response_class=JSONResponse)
  def generate(csrf_protect: CsrfProtect = Depends()):
    csrf_token, signed_token = csrf_protect.generate_csrf_tokens()
    response: JSONResponse = JSONResponse(
      status_code=200, content={"detail": "OK", "csrf_token": csrf_token}
    )
    csrf_protect.set_csrf_cookie(signed_token, response)
    return response

  @app.get("/protected", response_class=JSONResponse)
  async def protected(request: Request, csrf_protect: CsrfProtect = Depends()):
    await csrf_protect.validate_csrf(request)
    response: JSONResponse = JSONResponse(status_code=200, content={"detail": "OK"})
    csrf_protect.unset_csrf_cookie(response)  # prevent token reuse
    return response

  @app.exception_handler(CsrfProtectError)
  def csrf_protect_error_handler(request: Request, exc: CsrfProtectError):
    return JSONResponse(status_code=exc.status_code, content={"detail": exc.message})

  return TestClient(app)



#!/usr/bin/env python3
# Copyright (C) 2020-2024 All rights reserved.
# FILENAME:    ~~/tests/token_invalid.py
# VERSION:     0.3.7
# CREATED:     2020-11-26 16:14
# AUTHOR:      Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************
### Third-Party packages ###
from fastapi.testclient import TestClient
from warnings import filterwarnings

### Local modules ###
from . import test_client
from fastapi_csrf_protect import CsrfProtect


def test_validate_token_invalid_request(test_client: TestClient):
  @CsrfProtect.load_config
  def get_configs():
    return [("secret_key", "secret"), ("cookie-key", "fastapi-csrf-token")]

  ### Ignore DeprecationWarnings when setting cookie manually with FastAPI TestClient ###
  filterwarnings("ignore", category=DeprecationWarning)

  ### Get protected contents ###
  headers: dict = {"X-CSRF-Token": "invalid"}
  response = test_client.get(
    "/protected", cookies={"fastapi-csrf-token": "invalid"}, headers=headers
  )

  ### Assertions ###
  assert response.status_code == 401
  assert response.json() == {"detail": "The CSRF token is invalid."}



#!/usr/bin/env python3
# Copyright (C) 2020-2024 All rights reserved.
# FILENAME:    ~~/src/fastapi_csrf_protect/load_config.py
# VERSION:     0.3.4
# CREATED:     2020-11-25 14:35
# AUTHOR:      Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************
from typing import Literal, Optional, Set
from pydantic import (
  BaseModel,
  Field,
  StrictBool,
  StrictInt,
  StrictStr,
  model_validator,
)


class LoadConfig(BaseModel):
  cookie_key: Optional[StrictStr] = "fastapi-csrf-token"
  cookie_path: Optional[StrictStr] = "/"
  cookie_domain: Optional[StrictStr] = None
  cookie_samesite: Optional[Literal["lax", "none", "strict"]] = "lax"
  cookie_secure: Optional[StrictBool] = False
  header_name: Optional[StrictStr] = "X-CSRF-Token"
  header_type: Optional[StrictStr] = None
  httponly: Optional[StrictBool] = True
  max_age: Optional[StrictInt] = 3600
  methods: Optional[Set[Literal["DELETE", "GET", "OPTIONS", "PATCH", "POST", "PUT"]]] = Field(
    {"DELETE", "PATCH", "POST", "PUT"}, validate_default=True
  )
  secret_key: Optional[StrictStr] = None
  token_location: Optional[Literal["body", "header"]] = "header"
  token_key: Optional[StrictStr] = None

  @model_validator(mode="after")
  def validate_cookie_samesite_none_secure(self) -> "LoadConfig":
    if self.cookie_samesite in {None, "none"} and self.cookie_secure is not True:
      raise ValueError('The "cookie_secure" must be True if "cookie_samesite" set to "none".')
    return self

  @model_validator(mode="after")
  def validate_token_key(self) -> "LoadConfig":
    token_location: str = self.token_location if self.token_location is not None else "header"
    if token_location == "body":
      if self.token_key is None:
        raise ValueError('The "token_key" must be present when "token_location" is "body"')
    return self


__all__ = ("LoadConfig",)



#!/usr/bin/env python3
# Copyright (C) 2020-2024 All rights reserved.
# FILENAME:    ~~/src/fastapi_csrf_protect/exceptions.py
# VERSION:     0.3.7
# CREATED:     2020-11-25 14:35
# AUTHOR:      Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************
class CsrfProtectError(Exception):
  def __init__(self, status_code: int, message: str):
    self.status_code = status_code
    self.message = message


class InvalidHeaderError(CsrfProtectError):
  def __init__(self, message: str):
    super().__init__(422, message)


class MissingTokenError(CsrfProtectError):
  def __init__(self, message: str):
    super().__init__(400, message)


class TokenValidationError(CsrfProtectError):
  def __init__(self, message: str):
    super().__init__(401, message)


__all__ = ("CsrfProtectError", "InvalidHeaderError", "MissingTokenError", "TokenValidationError")



#!/usr/bin/env python3
# Copyright (C) 2021-2023 All rights reserved.
# FILENAME:  examples/form_data.py
# VERSION: 	 0.3.3
# CREATED: 	 2023-11-04 14:50
# AUTHOR: 	 Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************
from fastapi import FastAPI, Form, Request, Depends
from fastapi.responses import JSONResponse
from fastapi.templating import Jinja2Templates
from fastapi_csrf_protect import CsrfProtect
from fastapi_csrf_protect.exceptions import CsrfProtectError
from pydantic import BaseModel, EmailStr, StrictStr

app = FastAPI()
templates = Jinja2Templates(directory="templates")


class CsrfSettings(BaseModel):
  secret_key: str = "asecrettoeverybody"
  cookie_samesite: str = "none"
  cookie_secure: bool = True
  token_location: str = "body"
  token_key: str = "csrf-token"


@CsrfProtect.load_config
def get_csrf_config():
  return CsrfSettings()


@app.get("/")
async def form(request: Request, csrf_protect: CsrfProtect = Depends()):
  """
  Returns form template.
  """
  csrf_token, signed_token = csrf_protect.generate_csrf_tokens()
  response = templates.TemplateResponse("form.html", {"request": request, "csrf_token": csrf_token})
  csrf_protect.set_csrf_cookie(signed_token, response)
  return response


@app.post("/login", response_class=JSONResponse)
async def login(
  request: Request,
  email: EmailStr = Form(),
  name: StrictStr = Form(),
  password: StrictStr = Form(),
  csrf_protect: CsrfProtect = Depends(),
) -> JSONResponse:
  """
  Login using form data
  """
  await csrf_protect.validate_csrf(request)
  response: JSONResponse = JSONResponse(status_code=200, content={"detail": "OK"})
  csrf_protect.unset_csrf_cookie(response)  # prevent token reuse
  return response


@app.exception_handler(CsrfProtectError)
def csrf_protect_exception_handler(_: Request, exc: CsrfProtectError):
  return JSONResponse(status_code=exc.status_code, content={"detail": exc.message})



#!/usr/bin/env python3
# Copyright (C) 2020-2024 All rights reserved.
# FILENAME:    ~~/tests/missing_token.py
# VERSION:     0.3.7
# CREATED:     2020-11-26 16:14
# AUTHOR:      Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Third-party packages ###
from fastapi.testclient import TestClient

### Local packages ###
from . import test_client
from fastapi_csrf_protect import CsrfProtect


def test_validate_missing_cookie_token_request(test_client: TestClient):
  ### Loads config ###
  @CsrfProtect.load_config
  def get_secret_key():
    return [("secret_key", "secret")]

  ### Generate token ###
  response = test_client.get("/gen-token")
  csrf_token: str = response.json().get("csrf_token", None)
  headers: dict = {"X-CSRF-Token": csrf_token} if csrf_token is not None else {}

  ### Clear previously received cookies ###
  test_client.cookies = None  # type: ignore

  ### Get protected contents ###
  response = test_client.get("/protected", headers=headers)

  ### Assertions ###
  assert response.status_code == 400
  assert response.json() == {"detail": "Missing Cookie: `fastapi-csrf-token`."}


def test_validate_missing_header_token_request(test_client: TestClient):
  ### Loads Config ###
  @CsrfProtect.load_config
  def get_secret_key():
    return [("secret_key", "secret")]

  ### Get CSRF Tokens ###
  response = test_client.get("/gen-token")

  ### Get protected contents ###
  response = test_client.get("/protected")

  ### Assertions ###
  assert response.status_code == 422
  assert response.json() == {"detail": 'Bad headers. Expected "X-CSRF-Token" in headers'}



#!/usr/bin/env python3
# Copyright (C) 2020-2024 All rights reserved.
# FILENAME:    ~~/src/fastapi_csrf_protect/core.py
# VERSION:     0.3.7
# CREATED:     2020-11-25 14:35
# AUTHOR:      Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************
from fastapi.requests import Request
from fastapi.responses import Response
from fastapi_csrf_protect.csrf_config import CsrfConfig
from fastapi_csrf_protect.exceptions import (
  InvalidHeaderError,
  MissingTokenError,
  TokenValidationError,
)
from itsdangerous import BadData, SignatureExpired, URLSafeTimedSerializer
from hashlib import sha1
from re import match
from os import urandom
from pydantic import create_model
from starlette.datastructures import Headers
from typing import Any, Dict, Optional, Tuple
from warnings import warn


class CsrfProtect(CsrfConfig):
  def generate_csrf(self, secret_key: Optional[str] = None) -> Tuple[str, str]:
    """
    Deprecated. Please use `generate_csrf_tokens` method instead.

    ---
    :param secret_key: (Optional) the secret key used when generating tokens for users
    :type secret_key: (str | None) Defaults to None.
    """
    warn("This is deprecated; version=0.3.6", DeprecationWarning, stacklevel=2)
    return self.generate_csrf_tokens(secret_key)

  def generate_csrf_tokens(self, secret_key: Optional[str] = None) -> Tuple[str, str]:
    """
    Generate a CSRF token and a signed CSRF token using server's secret key to be stored in
    cookie. R

    ---
    :param secret_key: (Optional) the secret key used when generating tokens for users
    :type secret_key: (str | None) Defaults to None.
    """
    secret_key = secret_key or self._secret_key
    if secret_key is None:
      raise RuntimeError("A secret key is required to use CsrfProtect extension.")
    serializer = URLSafeTimedSerializer(secret_key, salt="fastapi-csrf-token")
    token = sha1(urandom(64)).hexdigest()
    signed = serializer.dumps(token)
    return token, signed

  def get_csrf_from_body(self, data: bytes) -> str:
    """
    Get token from the request body

    ---
    :param data: attached request body containing cookie data with configured `token_key`
    :type data: bytes
    """
    fields: Dict[str, Any] = {self._token_key: (str, "csrf-token")}
    Body = create_model("Body", **fields)
    content: str = '{"' + data.decode("utf-8").replace("&", '","').replace("=", '":"') + '"}'
    body = Body.model_validate_json(content)
    return body.model_dump()[self._token_key]

  def get_csrf_from_headers(self, headers: Headers) -> str:
    """
    Get token from the request headers

    ---
    :param headers: Headers containing header with configured `header_name`
    :type headers: starlette.datastructures.Headers
    """
    header_name, header_type = self._header_name, self._header_type
    header_parts = None
    try:
      header_parts = headers[header_name].split()
    except KeyError:
      raise InvalidHeaderError(f'Bad headers. Expected "{header_name}" in headers')
    token = None
    # Make sure the header is in a valid format that we are expecting, ie
    if not header_type:
      # <HeaderName>: <Token>
      if len(header_parts) != 1:
        raise InvalidHeaderError(f'Bad {header_name} header. Expected value "<Token>"')
      token = header_parts[0]
    else:
      # <HeaderName>: <HeaderType> <Token>
      if not match(r"{}\s".format(header_type), headers[header_name]) or len(header_parts) != 2:
        raise InvalidHeaderError(
          f'Bad {header_name} header. Expected value "{header_type} <Token>"'
        )
      token = header_parts[1]
    return token

  def set_csrf_cookie(self, csrf_signed_token: str, response: Response) -> None:
    """
    Sets Csrf Protection token to the response cookies

    ---
    :param csrf_signed_token: signed CSRF token from `generate_csrf_token` method
    :type csrf_signed_token: str
    :param response: The FastAPI response object to sets the access cookies in.
    :type response: fastapi.responses.Response
    """
    if not isinstance(response, Response):
      raise TypeError("The response must be an object response FastAPI")
    response.set_cookie(
      self._cookie_key,
      csrf_signed_token,
      max_age=self._max_age,
      path=self._cookie_path,
      domain=self._cookie_domain,
      secure=self._cookie_secure,
      httponly=self._httponly,
      samesite=self._cookie_samesite,
    )

  def unset_csrf_cookie(self, response: Response) -> None:
    """
    Remove Csrf Protection token from the response cookies

    ---
    :param response: The FastAPI response object to delete the access cookies in.
    :type response: fastapi.responses.Response
    """
    if not isinstance(response, Response):
      raise TypeError("The response must be an object response FastAPI")
    response.delete_cookie(self._cookie_key, path=self._cookie_path, domain=self._cookie_domain)

  async def validate_csrf(
    self,
    request: Request,
    cookie_key: Optional[str] = None,
    secret_key: Optional[str] = None,
    time_limit: Optional[int] = None,
  ):
    """
    Check if the given data is a valid CSRF token. This compares the given
    signed token to the one stored in the session.

    ---
    :param request: incoming Request instance
    :type request: fastapi.requests.Request
    :param cookie_key: (Optional) field name for the CSRF token field stored in cookies
        Default is set in CsrfConfig when `load_config` was called;
    :type cookie_key: str
    :param secret_key: (Optional) secret key used to decrypt the token
        Default is set in CsrfConfig when `load_config` was called;
    :type secret_key: str
    :param time_limit: (Optional) Number of seconds that the token is valid.
        Default is set in CsrfConfig when `load_config` was called;
    :type time_limit: int
    :raises TokenValidationError: Contains the reason that validation failed.
    """
    secret_key = secret_key or self._secret_key
    if secret_key is None:
      raise RuntimeError("A secret key is required to use CsrfProtect extension.")
    cookie_key = cookie_key or self._cookie_key
    signed_token = request.cookies.get(cookie_key)
    if signed_token is None:
      raise MissingTokenError(f"Missing Cookie: `{cookie_key}`.")
    time_limit = time_limit or self._max_age
    token: str
    if self._token_location == "header":
      token = self.get_csrf_from_headers(request.headers)
    else:
      if hasattr(request, "_json"):
        token = request._json.get(self._token_key, "")
      elif hasattr(request, "_form") and request._form is not None:
        token = request._form.get(self._token_key, "")
      else:
        token = self.get_csrf_from_body(await request.body())
    serializer = URLSafeTimedSerializer(secret_key, salt="fastapi-csrf-token")
    try:
      signature: str = serializer.loads(signed_token, max_age=time_limit)
      if token != signature:
        raise TokenValidationError("The CSRF signatures submitted do not match.")
    except SignatureExpired:
      raise TokenValidationError("The CSRF token has expired.")
    except BadData:
      raise TokenValidationError("The CSRF token is invalid.")


__all__ = ("CsrfProtect",)



#!/usr/bin/env python3
# Copyright (C) 2020-2024 All rights reserved.
# FILENAME:    ~~/src/fastapi_csrf_protect/__init__.py
# VERSION:     0.3.7
# CREATED:     2020-11-25 14:35
# AUTHOR:      Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************
"""
FastAPI extension that provides Csrf Protection Token support
"""

__version__ = "0.3.7"

from .core import CsrfProtect

__all__ = ("CsrfProtect",)



#!/usr/bin/env python3
# Copyright (C) 2020-2024 All rights reserved.
# FILENAME:    ~~/tests/load_config.py
# VERSION:     0.3.7
# CREATED:     2020-11-26 16:14
# AUTHOR:      Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Any

### Third-party packages ###
from pydantic import ValidationError
from pytest import mark

### Local modules ###
from fastapi_csrf_protect import CsrfProtect


@mark.parametrize(
  "config_key, config_value, valid",
  [
    ("header_name", 2, False),
    ("header_name", 1.0, False),
    ("header_name", True, False),
    ("header_name", b"header_name", False),
    ("header_name", "header_name", True),
    ("header_name", [], False),
    ("header_name", {}, False),
    ("header_type", 2, False),
    ("header_type", 1.0, False),
    ("header_type", True, False),
    ("header_type", b"header_type", False),
    ("header_type", "header_type", True),
    ("header_type", [], False),
    ("header_type", {}, False),
    ("methods", 2, False),
    ("methods", 1.0, False),
    ("methods", True, False),
    ("methods", b"GET, POST", False),
    ("methods", "GET, POST", False),
    ("methods", [], True),
    ("methods", {}, False),
    ("methods", [1, 2, 3], False),
    ("methods", (1, 2, 3), False),
    ("methods", {1, 2, 3}, False),
    ("methods", ["1", "2", "3"], False),
    ("methods", ("1", "2", "3"), False),
    ("methods", {"1", "2", "3"}, False),
    ("methods", ["GET", "POST", "DELETE"], True),
    ("methods", ("GET", "POST", "DELETE"), True),
    ("methods", {"GET", "POST", "DELETE"}, True),
    ("methods", {"key": "value"}, False),
    ("secret_key", 2, False),
    ("secret_key", 1.0, False),
    ("secret_key", True, False),
    ("secret_key", b"secret", False),
    ("secret_key", "secret", True),
    ("secret_key", [], False),
    ("secret_key", {}, False),
    ("token_location", "body", False),  # missing token_key
    ("token_location", b"body", False),
    ("token_location", "header", True),
    ("token_location", b"header", False),
    ("cookie_samesite", None, False),  # missing cookie_secure
    ("cookie_samesite", "lax", True),
    ("cookie_samesite", "none", False),  # missing cookie_secure
    ("cookie_samesite", "strict", True),
    ("cookie_samesite", b"lax", False),
    ("cookie_samesite", b"none", False),
    ("cookie_samesite", b"strict", False),
    ("cookie_samesite", "null", False),
    ("cookie_samesite", b"null", False),
    ("cookie_samesite", 0, False),
    ("cookie_samesite", 1, False),
    ("cookie_samesite", True, False),
    ("cookie_samesite", False, False),
    ("cookie_samesite", 2.0, False),
    ("cookie_samesite", {1, 2, 3}, False),
    ("cookie_samesite", {1.0, 2.0, 3.0}, False),
    ("cookie_samesite", {"1", "2", "3"}, False),
    ("cookie_samesite", [1, 2, 3], False),
    ("cookie_samesite", [1.0, 2.0, 3.0], False),
    ("cookie_samesite", ["1", "2", "3"], False),
    ("cookie_samesite", {"key": "value"}, False),
  ],
)
def test_load_config(config_key: str, config_value: Any, valid: bool):
  error_raised: bool = False
  try:

    @CsrfProtect.load_config
    def load_configs():
      return [(config_key, config_value)]

  except Exception as err:
    error_raised = True
    assert isinstance(err, ValidationError)
  assert error_raised is (True, False)[valid]



#!/usr/bin/env python3
# Copyright (C) 2020-2024 All rights reserved.
# FILENAME:    ~~/tests/missing_secret.py
# VERSION:     0.3.7
# CREATED:     2021-08-18 23:58
# AUTHOR:      Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Third-Party packages ###
from fastapi.testclient import TestClient
from pytest import raises

### Local modules ###
from . import test_client
from fastapi_csrf_protect import CsrfProtect


def test_validate_missing_secret_key(test_client: TestClient):
  with raises(RuntimeError) as err:

    @CsrfProtect.load_config
    def load_secret_key():
      return [("secret_key", None)]

    test_client.get("/gen-token")
  assert err.match("A secret key is required to use CsrfProtect extension.")
